/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Smart pointer management
*  SCOPE :              Smart pointer management
*  FILE :               smartp.hxx
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma interface
#endif
#ifndef __smartp_HXX
#define __smartp_HXX

/* IMPORTS *******************************************************************/

#ifndef __virtualmemory_HXX
#include "virtualmemory.hxx"
#endif

#ifndef __pagehdlmem_HXX
#include "pagehdlmem.hxx"
#endif

#ifndef __virtualdbserver_HXX
#include "virtualdbserver.hxx"
#endif

#ifndef _var_H
#include "var.h"
#endif

#ifndef _page_HXX
#include "page.hxx"
#endif

/*****************************************************************************/

#define yoo_declare(x)                                      \
class Ref<x> : public yoo_ObjectX {                         \
  public :                                                  \
  Ref() {}                                                  \
  Ref(OID oid) : yoo_ObjectX(oid) {} ;                      \
  Ref(x* e) : yoo_ObjectX(e) {} ;                           \
  x* operator->() const {                                   \
    PageHdlMem* pd ;                                        \
    return (x*)                                             \
      ((pd=(PageHdlMem*)YOO_VirtualMemory->_inVMPagePID[_oid])!=NULL? \
       pd->_page->object(_oid.objNo) :                                \
       YOO_VirtualServer->findPage(this->_oid)->object(_oid.objNo)) ; \
  }                                                         \
  x& operator*() const {                                    \
    return *this->operator->() ;                            \
  }                                                         \
  operator x*() const {                                     \
    return this->_oid==NULL_OID? NULL : this->operator->(); \
  }							    \
  class Ref<x>& operator=(x* obj) {                         \
    this->assign(obj); return *this;                        \
  }                                                         \
} ;

#define yoo_declare1(x, p1)                                 \
class Ref<x> : public yoo_ObjectX {                         \
  public :                                                  \
  Ref() {}                                                  \
  Ref(OID oid) : yoo_ObjectX(oid) {} ;                      \
  Ref(x* e) : yoo_ObjectX(e) {} ;                           \
  x* operator->() const {                                   \
    PageHdlMem* pd ;                                        \
    return (x*)                                             \
      ((pd=(PageHdlMem*)YOO_VirtualMemory->_inVMPagePID[_oid])!=NULL? \
       pd->_page->object(_oid.objNo) :                                \
       YOO_VirtualServer->findPage(this->_oid)->object(_oid.objNo)) ; \
  }                                                         \
  x& operator*() const {                                    \
    return *this->operator->() ;                            \
  }                                                         \
  operator x*() const {                                     \
    return this->_oid==NULL_OID? NULL : this->operator->(); \
  }							    \
  class Ref<x>& operator=(x* obj) {                         \
    this->assign(obj); return *this;                        \
  }                                                         \
  operator p1*() const { return (p1*)getObj() ; }           \
  operator Ref<p1>() const { return Ref<p1>(this->_oid) ; } \
} ;                                                         

#define yoo_declare2(x, p1, p2)                             \
class Ref<x> : public yoo_ObjectX {                         \
  public :                                                  \
  Ref() {}                                                  \
  Ref(OID oid) : yoo_ObjectX(oid) {} ;                      \
  Ref(x* e) : yoo_ObjectX(e) {} ;                           \
  x* operator->() const {                                   \
    PageHdlMem* pd ;                                        \
    return (x*)                                             \
      ((pd=(PageHdlMem*)YOO_VirtualMemory->_inVMPagePID[_oid])!=NULL? \
       pd->_page->object(_oid.objNo) :                                \
       YOO_VirtualServer->findPage(this->_oid)->object(_oid.objNo)) ; \
  }                                                         \
  x& operator*() const {                                    \
    return *this->operator->() ;                            \
  }                                                         \
  operator x*() const {                                     \
    return this->_oid==NULL_OID? NULL : this->operator->(); \
  }							    \
  class Ref<x>& operator=(x* obj) {                         \
    this->assign(obj); return *this;                        \
  }                                                         \
  operator p1*() const { return (p1*)getObj() ; }           \
  operator Ref<p1>() const { return Ref<p1>(this->_oid) ; } \
  operator p2*() const { return (p2*)getObj() ; }           \
  operator Ref<p2>() const { return Ref<p2>(this->_oid) ; } \
} ; 


class yoo_ObjectX {

  protected :
 
  OID _oid ;

  public :
  
  yoo_ObjectX() {}   
  yoo_ObjectX(void*); // Construtor from a pointer. 
  yoo_ObjectX(OID oid) : _oid(oid) {} ;
  OID& getOid() ; // Extract the encapsulated OID
  void assign(ObjectX* obj) ; // Assign from a pointer
  void assign(const yoo_ObjectX&) ; // Assign from another yoo_pointer
  ObjectX* getObj() const ;
  ObjectX* operator->() const ;
  ObjectX& operator*() const ;
  int operator==(ObjectX* p) const ;
  int operator==(const yoo_ObjectX& p) const ;
  int operator!=(ObjectX* p) const ;
  int operator!=(const yoo_ObjectX& p) const ; 
  int operator<(const yoo_ObjectX& p) const ; 
  int operator>(const yoo_ObjectX& p) const ;
  void setOid(OID& oid) { this->_oid = oid ; }
} ;

inline OID& yoo_ObjectX::getOid() {
  return this->_oid ;
}

inline ObjectX* yoo_ObjectX::operator->() const { 
  PageHdlMem* pd ;
  if ((pd = YOO_VirtualMemory->findPageHdl(_oid)) != NULL) {
    return pd->page()->object(_oid.objNo) ;
  } else {
    Page* p = YOO_VirtualServer->findPage(this->_oid) ;
    if (p == NULL) {
      return NULL ;
    } else {
      return p->object(_oid.objNo) ;
    }
  }
}

inline ObjectX& yoo_ObjectX::operator*() const { 
  return *(this->operator->()) ;
}

inline ObjectX* yoo_ObjectX::getObj() const { 
  return this->_oid == NULL_OID ? NULL : this->operator->() ;
}

inline int yoo_ObjectX::operator==(const yoo_ObjectX& p) const {
  return this->_oid == p._oid; 
}

inline int yoo_ObjectX::operator!=(ObjectX* p) const { 
  return ! (*this == p) ; 
}

inline int yoo_ObjectX::operator!=(const yoo_ObjectX& p) const {
  return ! (*this == p) ; 
}

inline int yoo_ObjectX::operator<(const yoo_ObjectX& p) const { 
  return (this->_oid < p._oid) ; 
}

inline int yoo_ObjectX::operator>(const yoo_ObjectX& p) const { 
  return (this->_oid > p._oid) ; 
}

#include "smartp_tmpl.hxx"

// Internal objects using yoo_pointers
yoo_declare(Cluster) 
yoo_declare(ObjectX) 

extern Ref<Cluster> YOO_ADMIN_CLU ;
extern Ref<Cluster> YOO_TMP_CLU ;
extern Ref<Cluster> YOO_DEF_CLU ;

#endif
